@@ -3,8 +3,29 @@ nitrogen.dfun
 -------------
 
 This module implements the class DFun, which 
-is used to wrap general differentiable, multi-variate
-functions.
+is used to interface general differentiable, multi-variate
+functions. See :doc:`tutorials/dfun` for an in-depth tutorial.
+
+
+===========================   ===================================
+**Create DFun objects from functions**
+-----------------------------------------------------------------
+:class:`DFun`                 The :class:`DFun` constructor
+:class:`FiniteDFun`           Finite-difference derivatives
+---------------------------   -----------------------------------
+**Modify single DFun objects**
+-----------------------------------------------------------------
+:class:`FixedInputDFun`       Fix input value(s).
+:class:`PermutedDFun`         Permute input and output ordering.
+:class:`ArrangedDFun`         Rearrange and duplicate outputs.
+---------------------------   -----------------------------------
+**Combine multiple DFun objects**
+-----------------------------------------------------------------
+:class:`CompositeDFun`        Function composition.
+:class:`MergedDFun`           Concatenate outputs.
+:class:`SimpleProduct`        Separable direct product.
+:class:`SelectedProduct`      Separable non-direct product.
+===========================   ===================================
 
 """
 
@@ -73,7 +94,7 @@ class DFun:
 
         Parameters
         ----------
-        X : ndarray
+        X : array_like
             An array of input values. `X` has shape (:attr:`nx`, ...).
         deriv : int, optional
             All derivatives up through order `deriv` are requested. The default is 0.
@@ -111,6 +132,7 @@ class DFun:
             raise ValueError('deriv is larger than maxderiv')
         
         # Check the shape of input X
+        X = np.array(X) 
         n = X.shape[0]
         if n != self.nx:
             raise ValueError(f'Expected input shape ({self.nx:d},...)')
@@ -282,13 +304,122 @@ class DFun:
             
         return out
     
-    def optimize(self, X0, fidx = 0, var = None, mode = 'min'):
+    def vj(self, X, var = None):
+        """
+        Wrapper for diff. function value and Jacobian
+
+        Parameters
+        ----------
+        X : ndarray
+            An (:attr:`nx`, ...) array of input values.
+        var : list of int
+            Variable list (see `var` in :func:`DFun.f`).
+            
+        Returns
+        -------
+        val : ndarray
+            The (nf,...) value.
+        jac : ndarray
+            The (nf,nv,...) Jacobian
+
+        """
+        
+        if var is None:
+            nvar = self.nx 
+        else:
+            nvar = len(var) 
+            
+        base_shape = X.shape[1:]
+        
+        d = self.f(X, deriv = 1, var = var) # Calculate 0th, 1st, and 2nd derivs
+        
+        #
+        # d has shape (nd, nf, ...)
+        #
+        
+        # Copy 0th derivatives to value
+        val = d[0,:].copy() 
+        
+        # Copy 1st derivatives to jacobian
+        # Notice swapping order of nd/nf from ndarray to jacobian
+        #
+        jac = np.ndarray( (self.nf, nvar,) + base_shape, dtype = d.dtype)
+        for i in range(self.nf):
+            np.copyto(jac[i,:], d[1:(nvar+1),i])
+     
+        return val, jac
+    
+    def vjh(self, X, var = None):
+        """
+        Wrapper for diff. function value, Jacobian, and Hessian
+
+        Parameters
+        ----------
+        X : ndarray
+            An (:attr:`nx`, ...) array of input values.
+        var : list of int
+            Variable list (see `var` in :func:`DFun.f`).
+            
+        Returns
+        -------
+        val : ndarray
+            The (nf,...) value.
+        jac : ndarray
+            The (nf,nv,...) Jacobian
+        hes : ndarray
+            The (nf,nv,nv,...) Hessian
+
+        """
+        
+        if var is None:
+            nvar = self.nx 
+        else:
+            nvar = len(var) 
+            
+        base_shape = X.shape[1:]
+        
+        d = self.f(X, deriv = 2, var = var) # Calculate 0th, 1st, and 2nd derivs
+        
+        #
+        # d has shape (nd, nf, ...)
+        #
+        
+        # Copy 0th derivatives to value
+        val = d[0,:].copy() 
+        
+        # Copy 1st derivatives to jacobian
+        # Notice swapping order of nd/nf from ndarray to jacobian
+        #
+        jac = np.ndarray( (self.nf, nvar,) + base_shape, dtype = d.dtype)
+        for i in range(self.nf):
+            np.copyto(jac[i,:], d[1:(nvar+1),i])
+        
+        # Copy 2nd derivatives to hessian
+        hes = np.ndarray( (self.nf, nvar, nvar) + base_shape, dtype = d.dtype)
+        
+        idx = nvar + 1 
+        for i in range(nvar):
+            for j in range(i,nvar):
+                # (i,j) derivative
+                
+                if i == j:
+                    np.copyto(hes[:,i,j], 2.0 * d[idx,:])
+                else:
+                    np.copyto(hes[:,i,j], d[idx,:])
+                    np.copyto(hes[:,j,i], d[idx,:])
+                    
+                idx = idx + 1
+        
+        return val, jac, hes 
+        
+    
+    def optimize(self, X0, fidx = 0, var = None, mode = 'min', disp = False):
         """
         Optimize an output value of a diff. function.
 
         Parameters
         ----------
-        X0 : ndarray
+        X0 : array_like
             (`nx`,) array containing the initial guess
         fidx : int 
             The DFun function index to optimize. The default is 0.
@@ -298,6 +429,8 @@ class DFun:
         mode : {'min'}, optional
             The optimization mode. 'min' determines the function
             minimum. The default is 'min'.
+        disp : bool
+            If True, messages will be displayed.
 
         Returns
         -------
@@ -308,6 +441,8 @@ class DFun:
 
         """
         
+        X0 = np.array(X0)
+        
         if var is None:
             var = [i for i in range(self.nx)]
         
@@ -334,7 +469,7 @@ class DFun:
         
         if mode == 'min':
             res = spopt.minimize(fun, X0[var], method = 'BFGS', 
-                                 jac = jac) #, hess = hess)
+                                 jac = jac, options = {'disp':disp} ) #, hess = hess)
         
             Xopt = X0.copy()
             Xopt[var] = res.x 
@@ -368,7 +503,7 @@ class DFun:
 
         """
         
-        # Calculate derivates of f to order deriv + 1
+        # Calculate derivatives of f to order deriv + 1
         F = self.f(X, deriv = deriv + 1, out = None, var = var)
         
         if var is None:
@@ -396,6 +531,18 @@ class DFun:
             
         return out
     
+    def _parse_out_var(self, X, deriv, out, var):
+        """ Parse a out and var, which may be None"""
+        if var is None:
+            var = [i for i in range(self.nx)]
+        if out is None:
+            nd, nvar = ndnvar(deriv, var, self.nx)
+            out = np.ndarray( (nd, self.nf) + X.shape[1:], dtype = X.dtype)
+        
+        return out, var
+        
+        
+    
 class CompositeDFun(DFun):
     
     """
@@ -864,7 +1011,399 @@ class MergedDFun(DFun):
     def __repr__(self):
         return f"MergedDFun({self.dfuns!r})"
            
+class PermutedDFun(DFun):
+    """
+    Permute the input or output order of a DFun
+    
+    Attributes
+    ----------
+    
+    df : DFun
+        The original DFun
+    in_order : ndarray
+        The new input order 
+    out_order : ndarray
+        The new output order
+    
+    """
+    
+    def __init__(self, df, in_order = None, out_order = None):
+        """
         
+
+        Parameters
+        ----------
+        df : DFun
+            A DFun object.
+        in_order : array_like of int, optional
+            The new input order. If None (default), the original
+            input order is used. For example, [2, 0, 1] moves
+            the third input first, the first input second, and the
+            second input last.
+        out_order : array_like of int, optional
+            The new output order. If None (default), the original
+            output order is used.        
+
+        """
+        
+        nx = df.nx 
+        nf = df.nf 
+        
+        if in_order is None:
+            in_order = np.arange(nx)
+        else:
+            in_order = np.array(in_order)
+            
+        if out_order is None:
+            out_order = np.arange(nf)
+        else:
+            out_order = np.array(out_order)
+            
+        if not np.all(np.sort(in_order) == np.arange(nx)):
+            raise ValueError('invalid in_order')
+        if not np.all(np.sort(out_order) == np.arange(nf)):
+            raise ValueError('invalid out_order') 
+        
+        
+        super().__init__(self._fpermute, nf, nx, df.maxderiv, df.zlevel)
+        
+        self.df = df
+        self.in_order = in_order
+        self.out_order = out_order 
+        
+        
+        # create the reverse direction in_order 
+        reverse_in = [None for i in range(nx)]
+        for i in range(nx):
+            reverse_in[in_order[i]] = i
+        
+        self._reverse_in = np.array(reverse_in)
+        
+        return
+
+    def _fpermute(self, X, deriv = 0, out = None, var = None):
+        
+        #
+        # X is in the input variables in the new order
+        # Permute them back to the old order first
+        #
+        X_old = X[self._reverse_in]
+        
+        #
+        # var is the requested variables refering to new
+        # input labels. Translate this to old input labels
+        #
+        if var is None:
+            var = [i for i in range(self.nx)]
+        
+        var_old = [self.in_order[v]  for v in var]
+        
+        # 
+        # Evaluate original dfun 
+        #
+        out_old = self.df.f(X_old, deriv = deriv, out = None, var = var_old) 
+        
+        #
+        # Finally, re-order the output values **in-place**
+        # 
+        
+        # Set up output
+        nd,nvar = ndnvar(deriv, var, self.nx)
+        if out is None:
+            base_shape = X.shape[1:]
+            out = np.ndarray( (nd, self.nf) + base_shape, dtype = X.dtype)
+            
+        # Copy old output to new order 
+        for i in range(self.nf):
+            iold = self.out_order[i]
+            np.copyto(out[:,i], out_old[:,iold])
+            
+        return out 
+    
+    def __repr__(self):
+        return f"PermutedDFun({self.df!r}, in_order = {self.in_order!r}, out_order = {self.out_order!r})"
+
+class ArrangedDFun(DFun):
+    """
+    Select a subset or duplicates of output functions.
+    
+    Attributes
+    ----------
+    
+    df : DFun
+        The original DFun
+    select : ndarray
+        The output functions in terms of the indices
+        of the original output functions.
+    
+    """
+    
+    def __init__(self, df, select):
+        """
+        
+        Parameters
+        ----------
+        df : DFun
+            A DFun object.
+        select : array_like of int
+            The new output functions. Repetitions are allowed.
+
+        """
+        
+        nx = df.nx 
+        
+        select = np.array(select, copy = True)
+        nf = len(select)
+            
+        if np.any(select < 0) or np.any(select >= df.nf):
+            raise ValueError("select has an element out of bounds")
+        
+        super().__init__(self._farrange, nf, nx, df.maxderiv, df.zlevel)
+        
+        self.df = df
+        self.select = select 
+        
+        return
+
+    def _farrange(self, X, deriv = 0, out = None, var = None):
+        
+        #
+        # Evaluate the complete set of functions
+        #
+        out_all = self.df.f(X, deriv = deriv, out = None, var = var) 
+        
+        # Set up output
+        nd,nvar = ndnvar(deriv, var, self.nx)
+        if out is None:
+            base_shape = X.shape[1:]
+            out = np.ndarray( (nd, self.nf) + base_shape, dtype = X.dtype)
+            
+        # Copy old output to new order 
+        for i in range(self.nf):
+            iold = self.select[i]
+            np.copyto(out[:,i], out_all[:,iold])
+            
+        return out 
+    
+    def __repr__(self):
+        return f"ArrangedDFun({self.df!r}, {self.select!r})"
+
+
+class SimpleProduct(DFun):
+    """
+    
+    Create a product function from 
+    mutually independent factors.
+    
+    .. math:
+       
+       f(x,y,\cdots) = g(x)h(y) \cdots
+       
+    Multi-dimensional factors may be supported in the future.
+    
+    """
+    
+    def __init__(self, factors):
+        """
+    
+        Parameters
+        ----------
+        factors : list of DFun or None
+            The factor for each input variable. Elements 
+            of None will be interpreted as unity.
+            
+        Notes
+        -----
+        If all elements are None, then 1 output variable
+        (equal to one) will be assumed. 
+
+        """
+        
+        nx = len(factors) # the number of factors 
+        
+        # Set the default parameters if all None
+        nf = 1 
+        maxderiv = None 
+        zlevel = 0  # Constant function (i.e. unity)
+        
+        for i in range(nx):
+            if factors[i] is not None: # There is some DFun presented
+                nf = factors[i].nf 
+                maxderiv = factors[i].maxderiv 
+                zlevel = factors[i].zlevel
+                break 
+        
+        # Now check all factors for agreement
+        for i in range(nx):
+            if factors[i] is not None:
+                if factors[i].nx != 1:
+                    raise ValueError("All factors must be 1-dimensional. This may"
+                                     " change in the future.")
+                if factors[i].nf != nf:
+                    raise ValueError("Factors do not have the same nf")
+                
+                maxderiv = _min_maxderiv(maxderiv, factors[i].maxderiv)
+                
+                if factors[i].zlevel is not None and factors[i].zlevel < 0 :
+                    zlevel = -1
+                elif zlevel is not None and zlevel < 0 :
+                    zlevel = -1 # remains zero 
+                else:
+                    zlevel = _sum_None(zlevel, factors[i].zlevel) 
+                
+        super().__init__(self._f_simple_prod, nf = nf, nx = nx,
+                         maxderiv = maxderiv, zlevel = zlevel) 
+        self.factors = factors 
+    
+    def _f_simple_prod(self, X, deriv = 0, out = None, var = None):
+        #
+        #
+        # Calculate derivative array for a product of
+        # independent functions
+        #  f(x1) * g(x2) * h(x3) * ...
+        #
+        factors = self.factors # The factors f, g, h, ...
+        
+        nd, nvar = ndnvar(deriv, var, self.nx) 
+        
+        out, var = self._parse_out_var(X, deriv, out, var)
+        
+        dv = []
+        for v in var:
+            if factors[v] is None:
+                # A unity factor
+                d = np.zeros((deriv+1, self.nf) + X.shape[1:], dtype = X.dtype)
+                d[0].fill(1.0) 
+            else:
+                # An explicit factor
+                d = factors[v].f(X[v:(v+1)], deriv) # (deriv+1, nf, ...)
+            dv.append(d)
+        #
+        # dv contains the 1-D derivatives of each factor in var respectively
+        # We need to combine these derivatives, and then scalar multiply
+        # by the values of non-var factors.
+        #
+        idxtab = adf.idxtab(deriv, nvar) 
+        for i in range(nd):
+            idx = idxtab[i]
+            out[i,:] = 1.0 
+            for j in range(nvar):
+                out[i,:] *= dv[j][idx[j]] 
+        
+        for i in range(self.nx):
+            if i not in var:
+                if factors[i] is None:
+                    pass # A factor of unity
+                else:
+                    val_i = factors[i].f(X[i:(i+1)], deriv = 0) # (1, nf, ...) 
+                    out *= val_i # broadcast onto (nd, nf, ...) with (1, nf, ...)
+        
+        return out
+
+class SelectedProduct(DFun):
+    
+    """
+    Given multiple sets of functions, create
+    a simple product using only selected combinatoins
+    of factors.
+    
+    """
+
+    def __init__(self, factors, select):
+        """
+        Parameters
+        ----------
+        factors : list of DFun
+            The factor for each input variable.
+        select : (n,len(factors)) array_like
+            The factor function selection indices.
+            
+        Notes
+        -----
+        Only 1D factors are currently supported. This may change 
+        in the future.
+
+        """
+        
+        nx = len(factors) # assume 1-D factors. this may change in the future
+        select = np.array(select).copy()  
+        nf = select.shape[0] # the number of selection products
+        if select.shape[1] != nx:
+            raise ValueError("select.shape[1] must equal the number of factors")
+        
+        maxderiv = None 
+        zlevel = 0  
+        
+        # Check all factors
+        for i in range(nx):
+            if factors[i].nx != 1:
+                raise ValueError("All factors must be 1-dimensional. This may"
+                                 " change in the future.")
+            if select[:,i].max() > (factors[i].nf - 1):
+                raise ValueError(f"factors[{i:d}] has too functions to select")
+                
+            maxderiv = _min_maxderiv(maxderiv, factors[i].maxderiv)
+            
+            if factors[i].zlevel is not None and factors[i].zlevel < 0 :
+                zlevel = -1
+            elif zlevel is not None and zlevel < 0 :
+                zlevel = -1 # remains zero 
+            else:
+                zlevel = _sum_None(zlevel, factors[i].zlevel) 
+                
+        super().__init__(self._f_select_prod, nf = nf, nx = nx,
+                         maxderiv = maxderiv, zlevel = zlevel) 
+        
+        self.factors = factors 
+        self.select = select 
+        
+    def _f_select_prod(self, X, deriv = 0, out = None, var = None):
+        #
+        #
+        # Calculate derivative array for a selected product of
+        # independent functions
+        #  f(x1) * g(x2) * h(x3) * ...
+        #
+        factors = self.factors # The factors f, g, h, ...
+        select = self.select   # The selected products 
+        
+        nd, nvar = ndnvar(deriv, var, self.nx) 
+        out, var = self._parse_out_var(X, deriv, out, var)
+        
+        dv = []
+        for v in var:
+            # (deriv+1, nf, ...)
+            dv.append(factors[v].f(X[v:(v+1)], deriv))
+            
+        # dv contains the 1-D derivatives of each factor in var respectively
+        # We need to combine these derivatives, and then scalar multiply
+        # by the values of non-var factors.
+        #
+        idxtab = adf.idxtab(deriv, nvar) 
+        
+        out.fill(1.0) # initialize all derivative products to 1
+        for i in range(nd):
+            idx = idxtab[i]
+            
+            for k in range(self.nf):
+                for j in range(nvar):
+                    #
+                    # For the j**th var, take the required
+                    # derivative of its respective factor
+                    #
+                    out[i,k] *= dv[j][idx[j], select[k,var[j]]] 
+        
+        # Now scale by value of non-var variables
+        for i in range(self.nx):
+            if i not in var:
+                val_i = factors[i].f(X[i:(i+1)], deriv = 0) # (1, nf, ...) 
+                
+                for k in range(self.nf):
+                    out[:,k] *= val_i[:,select[k,i]]
+                    # broadcast onto (nd, ...) with (1, ...)
+        
+        return out 
+    
 class PolyFactor(DFun):
     
     def __init__(self, terms, nx = None):
@@ -987,7 +1526,152 @@ class PolyPower(DFun):
         
         # Return result
         return adf2array([res],out)   
+
+class PowerExpansion(DFun):
     
+    """
+    Power series expansion about a given point. This is usually more efficient
+    than similar functions :class:`PolyPower` and :class:`PolyFactor`
+    because it uses derivative array translation instead of an explicit
+    sum over terms.
+    
+    Attributes
+    ----------
+    d : (nd,nf) ndarray
+        The defining derivative array about the expansion point
+    x0 : (nx,) ndarray
+        The expansion point. 
+    
+    """
+    
+    def __init__(self, d, x0):
+        """
+        Create a power series expansion 
+        
+        ..  math::
+            
+            f(\\mathbf{x}) = \\sum_\\alpha d^{(\\alpha)}(\\mathbf{x}-
+                                                        \\mathbf{x}_0)^\\alpha
+                
+        Parameters
+        ----------
+        d : array_like
+            The derivative array(s) at the expansion point.
+        x0 : array_like
+            The expansion point.
+            
+        """
+        
+        d = np.array(d) 
+        
+        if d.ndim == 1:
+            d = d.reshape((-1,1))
+        
+        # d now has shape (nd, nf)
+        
+        nd = d.shape[0]
+        nf = d.shape[1] 
+        
+        x0 = np.array(x0)
+        nx = len(x0)
+        
+        # Figure out the order of the expansion
+        #
+        order = 0
+        while True:
+            if nd == nderiv(order, nx):
+                # found
+                break
+            elif nd < nderiv(order, nx):
+                raise ValueError('the shape of d is inconsistent with x0')
+            else:
+                order += 1 
+        # order now equals the expansion order,
+        # which is equal to the zlevel of the DFun
+        
+        
+        super().__init__(self._fexpansion, nf = nf, nx = nx, maxderiv = None,
+                         zlevel = order)
+        
+        self.d = d 
+        self.x0 = x0 
+        self.order = self.zlevel 
+        self.idxD = adf.idxtab(self.order, self.nx)
+        self.nckD = adf.ncktab(self.order + self.nx + 1) 
+        
+        return 
+    
+    def _fexpansion(self, X, deriv = 0, out = None, var = None):
+
+        # Calculate the displacement from the 
+        # expansion origin
+        disp = np.empty_like(X)
+        for i in range(self.nx):
+            disp[i] = X[i] - self.x0[i] 
+        
+        # Now calculate the powers of the displacement
+        # [power, variable, ...]
+        disp_pow = np.ones((self.order+1,) + X.shape, dtype = X.dtype)
+        for k in range(self.order):
+            disp_pow[k+1] = disp_pow[k] * disp 
+
+        out, var = self._parse_out_var(X, deriv, out, var)
+        out.fill(0)
+        
+        nvar = len(var)
+        idx = adf.idxtab(deriv, nvar) # Index table for output derivative array
+        nd = nderiv(deriv, nvar)
+        
+        for iD in range(self.idxD.shape[0]):
+            
+            idxD = self.idxD[iD] # The index of the powers of d
+        
+            Dpow = np.ones(disp_pow.shape[2:], dtype = disp_pow.dtype)
+            for i in range(self.nx):
+                Dpow *= disp_pow[idxD[i], i]
+                
+            for iZ in range(nd):
+                idxZ = idx[iZ,:] # The result index in the requested
+                                 # `var` order
+                
+                idx_orig = np.arange(self.nx) * 0
+                for vid in range(nvar):
+                    idx_orig[var[vid]] = idxZ[vid]
+                # idx_orig is the result index in the complete variable order
+                
+                idxX = idxD + idx_orig # The index of the expansion coefficient
+                                       # that contributes to result
+                kX = np.sum(idxX)
+                if kX > self.order:
+                    break # Skip remaining 
+                
+                # Calculate the multi-index
+                # binomial coefficient
+                c = 1.0 
+                for i in range(self.nx):
+                    c *= self.nckD[idxX[i], min(idx_orig[i], idxD[i])]
+                    
+                iX = adf.idxpos(idxX, self.nckD)
+                
+                out[iZ,:] += c * self.d[iX,:] * Dpow
+        
+        return out
+
+def _min_maxderiv(maxA,maxB):
+    return _composite_maxderiv(maxA,maxB)
+def _max_zlevel(zlevelA,zlevelB):
+    return _merged_zlevel(zlevelA, zlevelB)
+def _product_None(a,b):
+    if a is None or b is None:
+        return None 
+    else:
+        return a * b
+def _sum_None(a,b):
+    if a is None or b is None:
+        return None 
+    else:
+        return a + b
+
 def _composite_maxderiv(maxA,maxB):
     """
     Determine the maxderiv value of a composite
@@ -1264,7 +1948,7 @@ def nderiv(deriv, nvar):
     k = min(deriv, nvar)
     return adf.ncktab(n,k)[n,k]
 
-def sym2invdet(S, deriv, nvar):
+def sym2invdet(S, deriv, nvar, logdet = False):
     """
     Calculate the inverse and determinant
     of a symmetric matrix. If S is real,
@@ -1282,6 +1966,9 @@ def sym2invdet(S, deriv, nvar):
         The maximum derivative order.
     nvar : int
         The number of independent variables.
+    logdet : boolean, optional
+        If True, the natural logarithm of the determinant
+        is returned instead. The default is False.
 
     Returns
     -------
@@ -1290,7 +1977,8 @@ def sym2invdet(S, deriv, nvar):
         in packed storage with shape (nd, nS, ...)
     det : ndarray
         The derivative array for det(S), with 
-        shape (nd, ...)
+        shape (nd, ...). (This equals ln(det(S)) if
+        `logdet` is True.)
 
     """
     
@@ -1320,17 +2008,36 @@ def sym2invdet(S, deriv, nvar):
     #    Stored in Spacked
     adf.chol_sp(Spacked, out = Spacked)
     
+    ############################################
     # 2) Compute the determinant
     #    This is the product of the squares of the diagonal entries
-    k = 0
-    for i in range(N):
-        if i == 0:
-            det = Spacked[k] 
-        else:
-            det = det * Spacked[k]
-        k = k + (i+2)
-    det = det * det
-    # det now equals the determinant of S
+    if logdet == False: # Calculate the normal determinant (default)
+        k = 0
+        for i in range(N):
+            if i == 0:
+                det = Spacked[k] 
+            else:
+                det = det * Spacked[k]
+            k = k + (i+2)
+        det = det * det
+        # det now equals the determinant of S
+        
+    else: # logdet == True, calculate the logarithm of the determinant instead 
+        # 
+        # The determinant is the product of the squares
+        # of the diagonal entries. The logarithm of the 
+        # determinant is thus twice the sum of
+        # logarithms of the diagonal entries.
+        k = 0
+        for i in range(N):
+            if i == 0:
+                det = adf.log(Spacked[k])
+            else:
+                det = det + adf.log(Spacked[k])
+            k = k + (i+2) 
+        det = 2.0 * det 
+    #
+    ############################################
     
     # 3) Compute the inverse of the Cholesky decomposition
     #    This overwrites Spacked
@@ -1345,8 +2052,8 @@ def sym2invdet(S, deriv, nvar):
         # S[:, i, ...] <-- Spacked[i].d
         np.copyto(iS[:,i], Spacked[i].d)
     
-    # S now contains the inverse of the original matrix
-    # 
-    # finally, return the ln(det) derivative array
+    # iS now contains the inverse of the original matrix
+    # and det the determinant (or log of det)
     #
-    return iS, det.d.copy()
\ No newline at end of file
+    return iS, det.d.copy()
+
