@@ -2,20 +2,22 @@
 nitrogen.special
 ----------------
 
-Special math and basis functions 
+Special math and basis functions implemented 
 as differentiable DFun objects.
 
-==============   ====================================================
-Function         Description
---------------   ----------------------------------------------------
-SinCosDFun       Real sine-cosine basis.
-LegendreLMCos    Associated Legendre functions with cosine argument.
-RealSphericalH   Real spherical harmonics.
-LaguerreL        Generalized Laguerre polynomials, :math:`L^{(\\alpha)}_n(x)`.
-RadialHO         Radial harmonic oscillator eigenfunctions in :math:`d` dimensions.
-BesselJ          Bessel functions of the first kind, :math:`J_{\\nu}(x)`.
-Real2DHO         Real 2-D isotropic harmonic oscillator wavefunctions.
-==============   ====================================================
+==============================================   ====================================================
+**Function**                                     **Description**
+----------------------------------------------   ----------------------------------------------------
+:class:`~nitrogen.special.SinCosDFun`            Real sine-cosine basis.
+:class:`~nitrogen.special.LegendreLMCos`         Associated Legendre functions with cosine argument.
+:class:`~nitrogen.special.RealSphericalH`        Real spherical harmonics.
+:class:`~nitrogen.special.LaguerreL`             Generalized Laguerre polynomials, :math:`L^{(\\alpha)}_n(x)`.
+:class:`~nitrogen.special.RadialHO`              Radial harmonic oscillator eigenfunctions in :math:`d` dimensions.
+:class:`~nitrogen.special.BesselJ`               Bessel functions of the first kind, :math:`J_{\\nu}(x)`.
+:class:`~nitrogen.special.Real2DHO`              Real 2-D isotropic harmonic oscillator wavefunctions.
+:class:`~nitrogen.special.Sin`                   Sine.
+:class:`~nitrogen.special.Monomial`              A monomial :math:`x^p y^q z^r \cdots`.
+==============================================   ====================================================
 
 """
 
@@ -36,16 +38,21 @@ class SinCosDFun(dfun.DFun):
        f_m(\\phi) &= 1/\sqrt{\pi} \sin(|m|\phi) &\ldots m < 0\\\\
        &= 1/\sqrt{2\pi}              &\ldots m = 0\\\\
        &= 1/\sqrt{\pi} \cos(m \phi)  &\ldots m > 0 
+       
+    If degree units are used, then the functions are renomalized
+    over :math:`[0^\\circ, 360^\\circ]`.
              
     Attributes
     ----------
     
     m : ndarray
         The m-index of each basis function (i.e. each output value).
+    angle : {'rad', 'deg'}
+        The angular units.
              
     """
     
-    def __init__(self, m):
+    def __init__(self, m, angle = 'rad'):
         """
         Create sine-cosine basis functions.
 
@@ -55,6 +62,8 @@ class SinCosDFun(dfun.DFun):
             If scalar, the :math:`2|m|+1` basis functions with
             index :math:`\leq |m|` will be included. If array_like,
             then `m` lists all m-indices to be included.
+        angle : {'rad', 'deg'}
+            The angle unit. The default is 'rad'. 
 
         """
         
@@ -65,11 +74,20 @@ class SinCosDFun(dfun.DFun):
             m = np.arange(-abs(m), abs(m)+1)
         else:
             m = np.array(m)
+            
+        if angle == 'rad':
+            rad = True
+        elif angle == 'deg':
+            rad = False
+        else:
+            raise ValueError('unexpected value of angle')
         
         super().__init__(self._fphi, nf = len(m), nx = 1,
                          maxderiv = None, zlevel = None)
         
         self.m = m 
+        self._rad = rad 
+        self.angle = angle 
         
         return 
     
@@ -83,7 +101,11 @@ class SinCosDFun(dfun.DFun):
             out = np.ndarray( (nd, self.nf) + base_shape, dtype = X.dtype)
             
         pi = np.pi
+        norm = pi if self._rad else 180.0
         phi = X 
+        
+        kfact = 1.0 # Running value of k! 
+        
         for k in range(nd):
             #
             # The k^th derivative order
@@ -92,25 +114,29 @@ class SinCosDFun(dfun.DFun):
                 m = self.m[i] # The m-index of the basis function
                 if m < 0:
                     # 1/sqrt(pi) * sin(|m| * phi) 
+                    
+                    m = abs(m) if self._rad else abs(m) * pi/180.0 
+                    
                     if k % 4 == 0:
-                        y = +abs(m)**k * np.sin(abs(m) * phi)
+                        y = +m**k * np.sin(m * phi)
                     elif k % 4 == 1:
-                        y = +abs(m)**k * np.cos(abs(m) * phi)
+                        y = +m**k * np.cos(m * phi)
                     elif k % 4 == 2:
-                        y = -abs(m)**k * np.sin(abs(m) * phi)
+                        y = -m**k * np.sin(m * phi)
                     else: # k % 4 == 3
-                        y = -abs(m)**k * np.cos(abs(m) * phi)
-                    np.copyto(out[k:(k+1),i],y / np.sqrt(pi))
+                        y = -m**k * np.cos(m * phi)
+                    np.copyto(out[k:(k+1),i],y / np.sqrt(norm))
                               
                 elif m == 0:
                     # 1/sqrt(2*pi)
                     if k == 0: # Zeroth derivative
-                        out[0:1,i].fill(1.0 / np.sqrt(2*pi))
+                        out[0:1,i].fill(1.0 / np.sqrt(2*norm))
                     if k > 0 : # All higher derivatives
                         out[k:(k+1),i].fill(0.0)
                         
                 elif m > 0 :
                     # 1/sqrt(pi) * cos(m * phi) 
+                    m = m if self._rad else m * pi/180.0 
                     if k % 4 == 0:
                         y = +m**k * np.cos(m * phi)
                     elif k % 4 == 1:
@@ -119,9 +145,12 @@ class SinCosDFun(dfun.DFun):
                         y = -m**k * np.cos(m * phi)
                     else: # k % 4 == 3
                         y = +m**k * np.sin(m * phi)
-                    np.copyto(out[k:(k+1),i],y / np.sqrt(pi))        
+                    np.copyto(out[k:(k+1),i],y / np.sqrt(norm))
+                    
+            out[k:(k+1)] /= kfact
+            kfact *= (k+1.0) # Update k! 
         
-        return out 
+        return out
 
 
 class LegendreLMCos(dfun.DFun):  
@@ -168,7 +197,7 @@ class LegendreLMCos(dfun.DFun):
         """
         
         if lmax < abs(m):
-            return ValueError("lmax must be >= |m|")
+            raise ValueError("lmax must be >= |m|")
         
         l = np.arange(abs(m), lmax+1)
         
@@ -309,6 +338,8 @@ class Sin(dfun.DFun):
             out = np.ndarray( (nd, self.nf) + base_shape, dtype = X.dtype)
             
         theta = X[0]
+        kfact = 1.0 
+        
         for k in range(nd):
             #
             # The k^th derivative order
@@ -322,7 +353,9 @@ class Sin(dfun.DFun):
             else: # k % 4 == 3
                 y = -np.cos(theta)   
                 
-            np.copyto(out[k,0:1], y)
+            np.copyto(out[k,0:1], y / kfact)
+            
+            kfact *= (k + 1.0) 
             
         return out 
     
@@ -356,7 +389,7 @@ class Monomial(dfun.DFun):
     
     def _monomial(self, X, deriv = 0, out = None, var = None):
         """ evaluation function """
-        raise NotImplementedError("NONO")
+        
         nd,nvar = dfun.ndnvar(deriv, var, self.nx)
         if out is None:
             base_shape = X.shape[1:]
@@ -367,7 +400,7 @@ class Monomial(dfun.DFun):
         for i in range(1, self.nx):
             res = res * adf.powi(x[i], self.pows[i])
             
-        dfun.adf2array(res, out)
+        dfun.adf2array([res], out)
         return out 
     
 class RealSphericalH(dfun.DFun):
@@ -577,9 +610,11 @@ class BesselJ(dfun.DFun):
         # multiple times. But this is not that expensive anyway,
         # so don't worry about the efficiency issue.
         #
+        kfact = 1.0 
         for k in range(nd):
-            dk = scipy.special.jvp(self.v, x, n = k)
+            dk = scipy.special.jvp(self.v, x, n = k) / kfact
             np.copyto(out[k:(k+1),0], dk)
+            kfact *= (k + 1.0) 
         
         return out 
     
@@ -869,7 +904,7 @@ class RadialHO(dfun.DFun):
         """
         
         if nmax < 0:
-            return ValueError("nmax must be >= 0")
+            raise ValueError("nmax must be >= 0")
          
         super().__init__(self._Rr, nf = nmax + 1, nx = 1,
                          maxderiv = None, zlevel = None)
@@ -926,7 +961,7 @@ class Real2DHO(dfun.DFun):
     ----------
     v : ndarray
         The :math:`v` quantum numbers, where
-        :math:`v = 2n + \\ell`.
+        :math:`v = 2n + \\vert \\ell \\vert`.
     ell : ndarray
         The :math:`\\ell` quantum numbers.
     n : ndarray
@@ -935,10 +970,12 @@ class Real2DHO(dfun.DFun):
         The initial `vmax` parameter.
     alpha : float
         The radial scaling parameter, :math:`\\alpha`.
+    angle : {'rad', 'deg'}
+        The angular unit. 
     
     """
     
-    def __init__(self, vmax, alpha, ell = None):
+    def __init__(self, vmax, alpha, ell = None, angle = 'rad'):
         """
         Create a 2D harmonic oscillator basis.
         
@@ -956,6 +993,8 @@ class Real2DHO(dfun.DFun):
             then `ell` lists all (signed) :math:`\\ell` values to be included.
             A value of None is equivalent to `ell` = `vmax`. The default is 
             None.
+        angle : {'rad', 'deg'}
+            The cylindrical angle unit. The default is 'rad'. 
 
         """
         
@@ -998,14 +1037,11 @@ class Real2DHO(dfun.DFun):
         # Save the alpha scaling parameter
         self.alpha = alpha 
         
-        # # Set up DFun's for the radial basis and the 
-        # # polar angular basis as separable factors 
-        # self.phi_basis = SinCosDFun(ell) # only supplies functions for index in `ell`
-        # radial_bases = []
-        # for ELL in ell: # Radial HO basis for each ELL in `ell`
-        #     nmax = round(np.floor(vmax-abs(ELL))/2)
-        #     radial_bases.append(RadialHO(nmax, ELL, d = 2, alpha = self.alpha))
-        # self.radial_bases = radial_bases
+        if angle == 'rad' or angle == 'deg':
+            self.angle = angle 
+            self._rad = True if angle == 'rad' else False 
+        else:
+            raise ValueError('unexpected angle unit') 
         
         
         return 
@@ -1021,7 +1057,7 @@ class Real2DHO(dfun.DFun):
         # Make adf objects for theta and phi 
         x = dfun.X2adf(X, deriv, var)
         r = x[0]        # Cylindrical radius
-        phi = x[1]      # Angular coordinate
+        phi = x[1] if self._rad else x[1] * np.pi/180.0      # Angular coordinate
         
         #########################################
         #
@@ -1039,13 +1075,22 @@ class Real2DHO(dfun.DFun):
         # Calculate the phi factors, f_ell(phi)
         sig_ell_uni = np.unique(self.ell) # list of unique signed ell
         fell = []
-        for sELL in sig_ell_uni:
-            if sELL == 0:
-                fell.append(adf.const_like(1/np.sqrt(2*np.pi), phi))
-            elif sELL < 0:
-                fell.append(1/np.sqrt(np.pi) * adf.sin(abs(sELL) * phi))
-            else: #sELL > 0
-                fell.append(1/np.sqrt(np.pi) * adf.cos(sELL * phi))
+        if self._rad: # Normalization assuming radians
+            for sELL in sig_ell_uni:
+                if sELL == 0:
+                    fell.append(adf.const_like(1/np.sqrt(2*np.pi), phi))
+                elif sELL < 0:
+                    fell.append(1/np.sqrt(np.pi) * adf.sin(abs(sELL) * phi))
+                else: #sELL > 0
+                    fell.append(1/np.sqrt(np.pi) * adf.cos(sELL * phi))
+        else: # Normalization for degrees
+            for sELL in sig_ell_uni:
+                if sELL == 0:
+                    fell.append(adf.const_like(1/np.sqrt(360.00), phi))
+                elif sELL < 0:
+                    fell.append(1/np.sqrt(180.0) * adf.sin(abs(sELL) * phi))
+                else: #sELL > 0
+                    fell.append(1/np.sqrt(180.0) * adf.cos(sELL * phi))
         #
         ##############################################
         
